package dbtypes

import (
	"fmt"
	"os"
	"path"

	"git.andr3h3nriqu3s.com/andr3/fyp/logic/db"
	"github.com/jackc/pgx/v5"
)

const (
	FAILED_TRAINING           = -4
	FAILED_PREPARING_TRAINING = -3
	FAILED_PREPARING_ZIP_FILE = -2
	FAILED_PREPARING          = -1

	PREPARING               = 1
	CONFIRM_PRE_TRAINING    = 2
	PREPARING_ZIP_FILE      = 3
	TRAINING                = 4
	READY                   = 5
	READY_ALTERATION        = 6
	READY_ALTERATION_FAILED = -6

	READY_RETRAIN        = 7
	READY_RETRAIN_FAILED = -7
)

type ModelHeadStatus int

const (
	MODEL_HEAD_STATUS_PRE_INIT ModelHeadStatus = 1
	MODEL_HEAD_STATUS_INIT                     = 2
	MODEL_HEAD_STATUS_TRAINING                 = 3
	MODEL_HEAD_STATUS_TRAINED                  = 4
	MODEL_HEAD_STATUS_READY                    = 5
)

type BaseModel struct {
	Name   string
	Status int
	Id     string

	ModelType    int    `db:"model_type"`
	ImageModeRaw string `db:"color_mode"`
	ImageMode    int    `db:"0"`
	Width        int
	Height       int
	Format       string
	CanTrain     int `db:"can_train"`
}

func GetBaseModel(db db.Db, id string) (base *BaseModel, err error) {
	var model BaseModel
	err = GetDBOnce(db, &model, "models where id=$1", id)
	if err != nil {
		return
	}
	model.ImageMode = StringToImageMode(model.ImageModeRaw)
	return &model, nil
}

func (m BaseModel) CanEval() bool {
	if m.Status != READY && m.Status != READY_RETRAIN && m.Status != READY_RETRAIN_FAILED && m.Status != READY_ALTERATION && m.Status != READY_ALTERATION_FAILED {
		return false
	}
	return true
}

func (m BaseModel) removeFailedDataPoints(c BasePack) (err error) {
	rows, err := c.GetDb().Query("select mdp.id from model_data_point as mdp join model_classes as mc on mc.id=mdp.class_id where mc.model_id=$1 and mdp.status=-1;", m.Id)
	if err != nil {
		return
	}
	defer rows.Close()

	base_path := path.Join("savedData", m.Id, "data")

	for rows.Next() {
		var dataPointId string
		err = rows.Scan(&dataPointId)
		if err != nil {
			return
		}

		p := path.Join(base_path, dataPointId+"."+m.Format)

		c.GetLogger().Warn("Removing image", "path", p)

		err = os.RemoveAll(p)
		if err != nil {
			return
		}
	}

	_, err = c.GetDb().Exec("delete from model_data_point as mdp using model_classes as mc where mdp.class_id = mc.id and mc.model_id=$1 and mdp.status=-1;", m.Id)
	return
}

// DO NOT Pass un filtered data on filters
func (m BaseModel) GetDefinitions(db db.Db, filters string, args ...any) ([]*Definition, error) {
	n_args := []any{m.Id}
	n_args = append(n_args, args...)
	return GetDbMultitple[Definition](db, fmt.Sprintf("model_definition as md where md.model_id=$1 %s", filters), n_args...)
}

// DO NOT Pass un filtered data on filters
func (m BaseModel) GetClasses(db db.Db, filters string, args ...any) ([]*ModelClass, error) {
	n_args := []any{m.Id}
	n_args = append(n_args, args...)
	return GetDbMultitple[ModelClass](db, fmt.Sprintf("model_classes as mc where mc.model_id=$1 %s", filters), n_args...)
}

type DataPointIterator struct {
	rows  pgx.Rows
	Model BaseModel
}

type DataPoint struct {
	Class int
	Path  string
}

func (iter DataPointIterator) Close() {
	iter.rows.Close()
}

func (m BaseModel) DataPoints(db db.Db, mode DATA_POINT_MODE) (data []DataPoint, err error) {
	rows, err := db.Query(
		"select mdp.id, mc.class_order, mdp.file_path from model_data_point as mdp inner "+
			"join model_classes as mc on mc.id = mdp.class_id "+
			"where mc.model_id = $1 and mdp.model_mode=$2;",
		m.Id, mode)
	if err != nil {
		return
	}
	defer rows.Close()

	data = []DataPoint{}

	for rows.Next() {
		var id string
		var class_order int
		var file_path string
		if err = rows.Scan(&id, &class_order, &file_path); err != nil {
			return
		}
		if file_path == "id://" {
			data = append(data, DataPoint{
				Path:  path.Join("./savedData", m.Id, "data", id+"."+m.Format),
				Class: class_order,
			})
		} else {
			panic("TODO remote file path")
		}
	}
	return
}

const RGB string = "rgb"
const GRAY string = "greyscale"

func StringToImageMode(colorMode string) int {
	switch colorMode {
	case GRAY:
		return 1
	case RGB:
		return 3
	default:
		panic("unkown color mode")
	}
}
